#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "config.h"

#ifndef HAVE_GETOPT
#include "getopt.c"
#endif

#define BLOCKSIZE 1024

int convertline(FILE * fileout, char *line, int force);

int main(int argc, char *argv[])
{
    int c;
    int pcorrupt = 0, force = 0;
    long int size = 0, slide = 0, corrupt = 0;
    char *line = NULL;
    char *outfile = NULL;
    int opt;
    extern int optind;
    extern char *optarg;
    FILE *fileout;
    FILE *filein;

    while ((opt = getopt(argc, argv, "o:hv")) != EOF) {
	switch (opt) {
	case 'o':
	    outfile = optarg;
	    break;

	case 'h':
	    printf("w3ctoelf [-o output-file] [-c] [-f] [files] ...\n\n-o output-file  File to write output to. Default is stdout.\n-c              Write corrupt lines to stderr.\n-f              Force conversion of corrupt lines. Garbage in, garbage out.\n");
	    exit(0);

	case 'v':
	    printf("w3ctoelf - " PACKAGE " " VERSION "\n");
	    exit(0);

	case 'c':
	    pcorrupt = 1;
	    break;

	case 'f':
	    force = 1;
	    break;

	}
    }

    /* open up file for output */
    if (outfile) {
	if (!(fileout = fopen(outfile, "w"))) {
	    fprintf(stderr, "File open error\n");
	    exit(1);
	}
    }
    else
	fileout = stdout;

    /* open up and read each file for input */
    opt = optind;
    if (opt == argc)
	filein = stdin;
    else
	filein = NULL;

    /* loop through each file */
    while (filein || opt < argc) {

	/* open the files for reading */
	if (!filein) {
	    if (!strcmp(argv[opt], "-"))
		filein = stdin;
	    else if (!(filein = fopen(argv[opt], "r"))) {
		perror(argv[opt]);
		exit(1);
	    }
	}

	/* get each line of text until EOF */
	while ((c = fgetc(filein)) != EOF) {
	    /* enough space in buffer? no, make some more */
	    if (slide >= size) {
		size += BLOCKSIZE;
		if (!(line = (char *) realloc(line, size))) {
		    fprintf(stderr, "Memory allocation error\n");
		    exit(1);
		}
	    }

	    /* save character to buffer */
	    line[slide++] = (char) c;

	    /* reached an end of line? if so, deal with the line */
	    if (c == 10) {

		/* the line may be terminated with a CR, if so chop it off
		 * otherwise replace the LF with a NULL to indicate end of string */
		if ((slide > 0) && (line[slide - 1] == 13))
		    line[slide - 2] = 0;
		else
		    line[slide - 1] = 0;

		/* is the line a comment? if so, ignore it */
		switch (line[0]) {
		case '#' :{
			if(!(strncmp(line,"#Fields: ",9))) convertline(fileout, line, force);
			else break;
		    }
		case 0:
		case '\t':
		case ' ':
		case ';':
		case '!':
		case '{':{
			break;
		    }
		default:{
			if (convertline(fileout, line, force)) {
			    if (pcorrupt)
				fprintf(stderr, "%s\n", line);
			    corrupt++;
			}
		    }
		}

		free(line);
		line = NULL;
		slide = 0;
		size = 0;
	    }

	}
	if (line)
	    free(line);

	/* close file (but only if the file was a real file) and loop */
	if (filein != stdin)
	    fclose(filein);
	filein = NULL;
	opt++;
    }

    /* close out output file */
    if (fileout != stdout)
	fclose(fileout);

    /* report corrupt line count */
    if (corrupt && !pcorrupt)
	fprintf(stderr, "%ld corrupt/incomplete lines ignored.\n", corrupt);

    exit(0);
}

/* get field number in a line */
/* buffer returned is size + 127 bytes so that it may be overwritten by the calling routine */
char *getfield(char *line, int field)
{
    char *buf;
    int start = 0, end = 0, slide = 0;
    int ws = 0;
    int br = 0;

    /* if ws is 0, we are finding text. if ws is not 0, we are finding whitespace */
    /* if br is not 0, we are inside a set of brackets */
    while (field && line[slide] && (line[slide] != 13)) {

	/* deal with brackets */
	if (line[slide] == '"') {
	    if (br)
		br = 0;
	    else
		br = 1;
	}

	/* find whitespace */
	switch (line[slide]) {
	case 8:
	case ' ':{
		if (!br && ws) {
		    ws = 0;
		}
		break;
	    }
	default:{
		if (!ws) {
		    field--;
		    ws = 1;
		    start = slide;
		}
	    }
	}
	slide++;
    }

    /* we found the start, now find the end */
    while (!end && line[slide] && (line[slide] != 13)) {

	/* find whitespace */
	switch (line[slide]) {
	case '"':{
		if (br)
		    br = 0;
		else
		    br = 1;
		break;
	    }
	case 8:
	case ' ':{
		if (!br) {
		    end = slide;
		}
		break;
	    }
	}
	slide++;
    }

    /* if !end, we found the last field */
    if (!end)
	end = slide;

    /* if field not found, return "" */
    if (field)
	start = end = 0;

    /* create a buffer for the field and return it - buffer contains an extra 127 bytes */
    if (!(buf = (char *) malloc(end - start + 128)))
	return (NULL);
    strncpy(buf, line + start, end - start);
    buf[end - start] = 0;
    return (buf);
}

void plustospace(char *str)
{
    register int x;

    for (x = 0; str[x]; x++)
	if (str[x] == '+')
	    str[x] = ' ';
}

/* convert the line as appropriate */
int sysdateno = 1;
int systimeno = 2;
int ipno = 3;
int authuserno = 4;
int methodno = 7;
int uristemno = 8;
int uriqueryno = 9;
int statusno = 10;
int versionno = 0;
int contentlengthno = 11;
int useragentno = 12;
int referrerno = 14;
int convertline(FILE * fileout, char *line, int force)
{
    int x;
    int bypass = 0;
    char *field;
    char *day;
    char *month;
    char *ip;
    char *authuser;
    char *sysdate;
    char *systime;
//    char *request;
    char *method;
    char *uristem;
    char *uriquery;
    char *status;
    char *version;
    char *contentlength;
    char *referrer;
    char *useragent;


    /* check if the line describes a line format - if so, parse that format
     * and then ignore that line */
    if (('#' == line[0])&&(!strncmp(line,"#Fields: ",9))) {

	/* reset defaults to blank */
	sysdateno = 0;
	systimeno = 0;
	ipno = 0;
	authuserno = 0;
	methodno = 0;
	uristemno = 0;
	uriqueryno = 0;
	statusno = 0;
	versionno = 0;
	contentlengthno = 0;
	useragentno = 0;
	referrerno = 0;

	x = 2;
	field = getfield(line,x);
	while(field[0]) {
	    if(!strcasecmp(field,"date")) sysdateno = x-1;
	    if(!strcasecmp(field,"time")) systimeno = x-1;
	    if(!strcasecmp(field,"c-ip")) ipno = x-1;
	    if(!strcasecmp(field,"cs-username")) authuserno = x-1;
	    if(!strcasecmp(field,"cs-method")) methodno = x-1;
	    if(!strcasecmp(field,"cs-uri-stem")) uristemno = x-1;
	    if(!strcasecmp(field,"cs-uri-query")) uriqueryno = x-1;
	    if(!strcasecmp(field,"sc-status")) statusno = x-1;
	    if(!strcasecmp(field,"cs-version")) versionno = x-1;
	    if(!strcasecmp(field,"sc-bytes")) contentlengthno = x-1;
	    if(!strcasecmp(field,"cs(user-agent)")) useragentno = x-1;
	    if(!strcasecmp(field,"cs(Referer)")) referrerno = x-1;

	    free(field);
	    field = getfield(line,++x);
	}
	free(field);
	return(0);
    }


    /* output ELF from W3C input */

    sysdate = getfield(line, sysdateno);
    systime = getfield(line, systimeno);
    ip = getfield(line, ipno);
    authuser = getfield(line, authuserno);
    method = getfield(line, methodno);
    uristem = getfield(line, uristemno);
    uriquery = getfield(line, uriqueryno);
    status = getfield(line, statusno);
    version = getfield(line, versionno);
    contentlength = getfield(line, contentlengthno);
    useragent = getfield(line, useragentno);
    plustospace(useragent);
    referrer = getfield(line, referrerno);


    /* check for memory errors */
    if (!sysdate || !systime || !ip || !authuser || !method || !uristem || !uriquery || !status || !version || !contentlength || !useragent || !referrer) {
	fprintf(stderr, "Memory allocation error\n");
	exit(1);
    }

    /* sanity check */
    if(!sysdateno) bypass = 1;	/* no date - these lines are bogus */
    if(!systimeno) bypass = 1;	/* no time - these lines are bogus */
    if(!ipno) bypass = 1;	/* no ip address - these lines are bogus */
    if(!authuserno) strcpy(authuser, "-");	/* no authuser, put in a dash */
    if(!methodno) bypass = 1;	/* no method, these lines are bogus */
    if(!uristemno) bypass = 1;	/* no uri stem, these lines are bogus */
    if(!uriqueryno) bypass = 1;	/* no uri query, these lines are bogus */
    if(!statusno) bypass = 1;	/* no status, these lines are bogus */
    if(!versionno) strcpy(version, "HTTP/1.0");	/* no version, so guess */
    if(!contentlengthno) strcpy(contentlength, "0");	/* no content length, so insert a zero */
    if(!useragentno) strcpy(useragent, "");	/* no useragent, so ignore */
    if(!referrerno) strcpy(referrer, "");	/* no referrer, so ignore */

    /* extended log format */
    if (strlen(sysdate) >= 10) {
	switch (sysdate[5]) {
	case '0':{
		switch (sysdate[6]) {
		case '1':{
			month = "Jan";
			break;
		    }
		case '2':{
			month = "Feb";
			break;
		    }
		case '3':{
			month = "Mar";
			break;
		    }
		case '4':{
			month = "Apr";
			break;
		    }
		case '5':{
			month = "May";
			break;
		    }
		case '6':{
			month = "Jun";
			break;
		    }
		case '7':{
			month = "Jul";
			break;
		    }
		case '8':{
			month = "Aug";
			break;
		    }
		case '9':{
			month = "Sep";
			break;
		    }
		}
		break;
	    }
	case '1':{
	    }
	    switch (sysdate[6]) {
	    case '0':{
		    month = "Oct";
		    break;
		}
	    case '1':{
		    month = "Nov";
		    break;
		}
	    case '2':{
		    month = "Dec";
		    break;
		}
	    }
	}
	day = sysdate + 8;
    }
    else {
	month = "";
	day = "";
    }

    /* output log line */
    if (!bypass || force) {
	fprintf(fileout, "%s - %s [%.2s/%s/%.4s:%s +0000] \"%s %s", ip, authuser, day, month, sysdate, systime, method, uristem);
	if ((uriquery[0] != '-') && (strlen(uriquery) > 1))
	    fprintf(fileout, "?%s", uriquery);
	fprintf(fileout, " %s\" %s %s", version, status, contentlength);
	
	if (referrer[0]) fprintf(fileout, " \"%s\"", referrer);
	if (useragent[0]) fprintf(fileout, " \"%s\"",  useragent);
	
	fprintf(fileout,"\n");
    }

    free(sysdate);
    free(systime);
    free(ip);
    free(authuser);
    free(method);
    free(uristem);
    free(uriquery);
    free(status);
    free(version);
    free(contentlength);
    free(useragent);
    free(referrer);

    return(bypass);
}
